/*#######################################################
 *
 *   Maintained by Gregor Santner, 2017-
 *   https://gsantner.net/
 *
 *   License of this file: Apache 2.0 (Commercial upon request)
 *     https://www.apache.org/licenses/LICENSE-2.0
 *
#########################################################*/
package net.gsantner.markor.activity;

import android.animation.Animator;
import android.animation.AnimatorListenerAdapter;
import android.annotation.SuppressLint;
import android.app.Activity;
import android.content.Context;
import android.graphics.Color;
import android.graphics.Typeface;
import android.os.Build;
import android.os.Bundle;
import android.text.TextUtils;
import android.util.TypedValue;
import android.view.Gravity;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.SubMenu;
import android.view.View;
import android.view.ViewGroup;
import android.webkit.JavascriptInterface;
import android.webkit.WebSettings;
import android.webkit.WebView;
import android.widget.HorizontalScrollView;
import android.widget.SearchView;
import android.widget.TextView;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.appcompat.app.AppCompatActivity;

import net.gsantner.markor.BuildConfig;
import net.gsantner.markor.R;
import net.gsantner.markor.format.TextConverter;
import net.gsantner.markor.format.TextFormat;
import net.gsantner.markor.format.general.DatetimeFormatDialog;
import net.gsantner.markor.model.Document;
import net.gsantner.markor.ui.AttachImageOrLinkDialog;
import net.gsantner.markor.ui.DraggableScrollbarScrollView;
import net.gsantner.markor.ui.FileInfoDialog;
import net.gsantner.markor.ui.FilesystemViewerCreator;
import net.gsantner.markor.ui.SearchOrCustomTextDialogCreator;
import net.gsantner.markor.ui.hleditor.HighlightingEditor;
import net.gsantner.markor.ui.hleditor.TextActions;
import net.gsantner.markor.util.AppSettings;
import net.gsantner.markor.util.ContextUtils;
import net.gsantner.markor.util.MarkorWebViewClient;
import net.gsantner.markor.util.ShareUtil;
import net.gsantner.opoc.activity.GsFragmentBase;
import net.gsantner.opoc.android.dummy.TextWatcherDummy;
import net.gsantner.opoc.net.OpocWebViewChromeClient;
import net.gsantner.opoc.preference.FontPreferenceCompat;
import net.gsantner.opoc.ui.FilesystemViewerData;
import net.gsantner.opoc.util.ActivityUtils;
import net.gsantner.opoc.util.CoolExperimentalStuff;
import net.gsantner.opoc.util.StringUtils;
import net.gsantner.opoc.util.TextViewUndoRedo;

import java.io.File;

@SuppressWarnings({"UnusedReturnValue"})
@SuppressLint("NonConstantResourceId")
public class DocumentEditFragment extends GsFragmentBase implements TextFormat.TextFormatApplier {
    public static final String FRAGMENT_TAG = "DocumentEditFragment";
    public static final String SAVESTATE_DOCUMENT = "DOCUMENT";
    public static final String START_PREVIEW = "START_PREVIEW";

    public static DocumentEditFragment newInstance(final @NonNull Document document, final Integer lineNumber, final boolean preview) {
        DocumentEditFragment f = new DocumentEditFragment();
        Bundle args = new Bundle();
        args.putSerializable(Document.EXTRA_DOCUMENT, document);
        if (lineNumber != null) {
            args.putInt(Document.EXTRA_FILE_LINE_NUMBER, lineNumber);
        }
        args.putBoolean(START_PREVIEW, preview);
        f.setArguments(args);
        return f;
    }

    public static DocumentEditFragment newInstance(final @NonNull File path, final Integer lineNumber) {
        return newInstance(new Document(path), lineNumber, false);
    }

    private HighlightingEditor _hlEditor;
    private ViewGroup _textActionsBar;
    private TextView _textSdWarning;
    private WebView _webView;
    private DraggableScrollbarScrollView _primaryScrollView;

    private AppSettings _appSettings;
    private HorizontalScrollView _hsView;
    private SearchView _menuSearchViewForViewMode;
    private Document _document;
    private TextFormat _textFormat;
    private ShareUtil _shareUtil;
    private TextViewUndoRedo _editTextUndoRedoHelper;
    private boolean _isPreviewVisible;
    private MarkorWebViewClient _webViewClient;
    private OpocWebViewChromeClient _webChromeClient;
    private boolean _nextConvertToPrintMode = false;
    private long _loadModTime = 0;
    private MenuItem _saveMenuItem, _undoMenuItem, _redoMenuItem;

    // Wrap text setting and wrap text state are separated as the wrap text state may depend on
    // if the file is in the main activity (quicknote and todotxt). Documents in mainactivity
    // will _always_ open wrapped, but can be explicitly be set to unwrapped through the menu.
    // Toggling the wrap state option will set and save the new value, but the file will always
    // open wrapped in the main activity.
    private boolean wrapTextSetting;
    private boolean wrapText;
    private boolean highlightText;
    private boolean autoFormat;

    public DocumentEditFragment() {
        super();
    }

    @Override
    public void onCreate(final Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        final Bundle args = getArguments();
        // Create the document as soon as possible
        if (savedInstanceState != null && savedInstanceState.containsKey(SAVESTATE_DOCUMENT)) {
            _document = (Document) savedInstanceState.getSerializable(SAVESTATE_DOCUMENT);
        } else if (args != null && args.containsKey(Document.EXTRA_DOCUMENT)) {
            _document = (Document) args.get(Document.EXTRA_DOCUMENT);
        } else {
            // We must have a document
            _document = Document.getDefault(getContext());
        }
    }

    @Override
    protected int getLayoutResId() {
        return R.layout.document__fragment__edit;
    }

    @SuppressLint({"SetJavaScriptEnabled", "WrongConstant", "AddJavascriptInterface"})
    @Override
    public void onViewCreated(@NonNull View view, Bundle savedInstanceState) {
        super.onViewCreated(view, savedInstanceState);
        final Activity activity = getActivity();
        _appSettings = new AppSettings(activity);
        if (_appSettings.getSetWebViewFulldrawing() && Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP) {
            WebView.enableSlowWholeDocumentDraw();
        }

        _hlEditor = view.findViewById(R.id.document__fragment__edit__highlighting_editor);
        _textActionsBar = view.findViewById(R.id.document__fragment__edit__text_actions_bar);
        _textSdWarning = view.findViewById(R.id.document__fragment__edit__content_editor__permission_warning);
        _webView = view.findViewById(R.id.document__fragment_view_webview);
        _primaryScrollView = view.findViewById(R.id.document__fragment__edit__content_editor__scrolling_parent);

        _shareUtil = new ShareUtil(activity);

        _webViewClient = new MarkorWebViewClient(activity);
        _webChromeClient = new OpocWebViewChromeClient(_webView, getActivity(), view.findViewById(R.id.document__fragment_fullscreen_overlay));
        _webView.setWebChromeClient(_webChromeClient);
        _webView.setWebViewClient(_webViewClient);
        _webView.addJavascriptInterface(this, "Android");
        WebSettings webSettings = _webView.getSettings();
        webSettings.setBuiltInZoomControls(true);
        webSettings.setDisplayZoomControls(false);
        webSettings.setTextZoom((int) (_appSettings.getViewFontSize() / 15.7f * 100f));
        webSettings.setCacheMode(WebSettings.LOAD_CACHE_ELSE_NETWORK);
        webSettings.setDatabaseEnabled(true);
        webSettings.setGeolocationEnabled(false);
        webSettings.setJavaScriptEnabled(true);
        webSettings.setDomStorageEnabled(true);
        webSettings.setAllowFileAccess(true);
        webSettings.setAllowContentAccess(true);
        webSettings.setAllowFileAccessFromFileURLs(true);
        webSettings.setAllowUniversalAccessFromFileURLs(true);
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.JELLY_BEAN_MR1) {
            webSettings.setMediaPlaybackRequiresUserGesture(false);
        }

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.KITKAT && BuildConfig.IS_TEST_BUILD && BuildConfig.DEBUG) {
            WebView.setWebContentsDebuggingEnabled(true); // Inspect on computer chromium browser: chrome://inspect/#devices
        }

        // Upon construction, the document format has been determined from extension etc
        // Here we replace it with the last saved format.
        _document.setFormat(_appSettings.getDocumentFormat(_document.getPath(), _document.getFormat()));
        applyTextFormat(_document.getFormat());
        _textFormat.getTextActions().setDocument(_document);

        if (activity instanceof DocumentActivity) {
            ((DocumentActivity) activity).setDocumentTitle(_document.getTitle());
        }

        _hlEditor.setScrollView(_primaryScrollView);
        _hlEditor.setLineSpacing(0, _appSettings.getEditorLineSpacing());

        _hlEditor.setTextSize(TypedValue.COMPLEX_UNIT_SP, _appSettings.getDocumentFontSize(_document.getPath()));
        _hlEditor.setTypeface(FontPreferenceCompat.typeface(getContext(), _appSettings.getFontFamily(), Typeface.NORMAL));

        _hlEditor.setBackgroundColor(_appSettings.getEditorBackgroundColor());
        _hlEditor.setTextColor(_appSettings.getEditorForegroundColor());

        _hlEditor.setGravity(_appSettings.isEditorStartEditingInCenter() ? Gravity.CENTER : Gravity.NO_GRAVITY);

        // Do not need to send contents to accessibility
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            _hlEditor.setImportantForAccessibility(View.IMPORTANT_FOR_AUTOFILL_NO_EXCLUDE_DESCENDANTS);
        }

        _webView.setBackgroundColor(Color.TRANSPARENT);
        final Bundle args = getArguments();
        setViewModeVisibility(args.getBoolean(START_PREVIEW, _appSettings.getDocumentPreviewState(_document.getPath())));

        _loadModTime = 0; // force next load
        loadDocument();

        _editTextUndoRedoHelper = new TextViewUndoRedo(_hlEditor);

        // Set initial wrap state
        initDocState();

        final Runnable debounced = StringUtils.makeDebounced(500, () -> {
            checkTextChangeState();
            updateUndoRedoIconStates();
        });
        _hlEditor.addTextChangedListener(TextWatcherDummy.after(s -> debounced.run()));
    }

    @Override
    public void onFragmentFirstTimeVisible() {

        int startPos = _appSettings.getLastEditPosition(_document.getPath(), _hlEditor.length());
        final Bundle args = getArguments();

        // First start - overwrite start position if needed
        if (_savedInstanceState == null) {
            if (isDisplayedAtMainActivity() || _appSettings.isEditorStartOnBotttom()) {
                startPos = _hlEditor.length();
            } else if (args != null && args.containsKey(Document.EXTRA_FILE_LINE_NUMBER)) {
                final int lno = args.getInt(Document.EXTRA_FILE_LINE_NUMBER);
                if (lno >= 0) {
                    startPos = StringUtils.getIndexFromLineOffset(_hlEditor.getText(), lno, 0);
                } else {
                    startPos = _hlEditor.length();
                }
            }
        }

        StringUtils.setSelectionAndShow(_hlEditor, startPos);
    }

    public void resume() {
        loadDocument();
        if (_hlEditor != null) {
            _hlEditor.onResume();
        }
        if (_webView != null) {
            _webView.onResume();
        }
    }

    @Override
    public void onResume() {
        resume();
        super.onResume();
    }

    @Override
    public void onSaveInstanceState(@NonNull Bundle outState) {
        outState.putSerializable(SAVESTATE_DOCUMENT, _document);
        super.onSaveInstanceState(outState);
    }

    public void pause() {
        saveDocument(false);
        if (_hlEditor != null) {
            _hlEditor.onPause();
        }
        if (_webView != null) {
            _webView.onPause();
        }

        if (_appSettings != null && _document != null) {
            _appSettings.addRecentDocument(_document.getFile());
            _appSettings.setDocumentPreviewState(_document.getPath(), _isPreviewVisible);
            if (_hlEditor != null) {
                _appSettings.setLastEditPosition(_document.getPath(), _hlEditor.getSelectionStart());
            }
        }
    }

    @Override
    public void onPause() {
        pause();
        super.onPause();
    }

    @Override
    public void onCreateOptionsMenu(Menu menu, MenuInflater inflater) {
        super.onCreateOptionsMenu(menu, inflater);
        inflater.inflate(R.menu.document__edit__menu, menu);
        ContextUtils cu = ContextUtils.get();
        cu.tintMenuItems(menu, true, Color.WHITE);
        cu.setSubMenuIconsVisiblity(menu, true);

        menu.findItem(R.id.action_undo).setVisible(_appSettings.isEditorHistoryEnabled());
        menu.findItem(R.id.action_redo).setVisible(_appSettings.isEditorHistoryEnabled());
        menu.findItem(R.id.action_send_debug_log).setVisible(MainActivity.IS_DEBUG_ENABLED && !isDisplayedAtMainActivity() && !_isPreviewVisible);

        final boolean isExperimentalFeaturesEnabled = _appSettings.isExperimentalFeaturesEnabled();

        // Undo / Redo / Save (keep visible, but deactivated and tinted grey if not executable)
        _undoMenuItem = menu.findItem(R.id.action_undo).setVisible(!_isPreviewVisible);
        _redoMenuItem = menu.findItem(R.id.action_redo).setVisible(!_isPreviewVisible);
        _saveMenuItem = menu.findItem(R.id.action_save).setVisible(!_isPreviewVisible);

        // Edit / Preview switch
        menu.findItem(R.id.action_edit).setVisible(_isPreviewVisible);
        menu.findItem(R.id.submenu_attach).setVisible(false);
        menu.findItem(R.id.action_preview).setVisible(!_isPreviewVisible);
        menu.findItem(R.id.action_search).setVisible(!_isPreviewVisible);
        menu.findItem(R.id.action_search_view).setVisible(_isPreviewVisible);
        menu.findItem(R.id.submenu_format_selection).setVisible(!_isPreviewVisible);

        menu.findItem(R.id.action_share_pdf).setVisible(Build.VERSION.SDK_INT >= Build.VERSION_CODES.KITKAT);
        menu.findItem(R.id.action_share_image).setVisible(true);
        menu.findItem(R.id.action_load_epub).setVisible(isExperimentalFeaturesEnabled);

        // SearchView (View Mode)
        _menuSearchViewForViewMode = (SearchView) menu.findItem(R.id.action_search_view).getActionView();
        _menuSearchViewForViewMode.setSubmitButtonEnabled(true);
        _menuSearchViewForViewMode.setQueryHint(getString(R.string.search));
        _menuSearchViewForViewMode.setOnQueryTextFocusChangeListener((v, searchHasFocus) -> {
            if (!searchHasFocus) {
                _menuSearchViewForViewMode.setQuery("", false);
                _menuSearchViewForViewMode.setIconified(true);
            }
        });
        _menuSearchViewForViewMode.setOnQueryTextListener(new SearchView.OnQueryTextListener() {
            @Override
            public boolean onQueryTextSubmit(String text) {
                _webView.findNext(true);
                return true;
            }

            @Override
            public boolean onQueryTextChange(String text) {
                _webView.findAllAsync(text);
                return true;
            }
        });

        // Set various initial states
        updateMenuToggleStates(_document.getFormat());
        checkTextChangeState();
        updateUndoRedoIconStates();
    }

    private void updateUndoRedoIconStates() {
        if (_editTextUndoRedoHelper == null) {
            return;
        }

        final boolean canUndo = _editTextUndoRedoHelper.getCanUndo();
        if (_undoMenuItem != null && _undoMenuItem.isEnabled() != canUndo) {
            _undoMenuItem.setEnabled(canUndo).getIcon().mutate().setAlpha(canUndo ? 255 : 40);
        }

        final boolean canRedo = _editTextUndoRedoHelper.getCanRedo();
        if (_redoMenuItem != null && _redoMenuItem.isEnabled() != canRedo) {
            _redoMenuItem.setEnabled(canRedo).getIcon().mutate().setAlpha(canRedo ? 255 : 40);
        }
    }

    public void loadDocument() {
        //Only trigger the load process if constructing or file updated
        final long modTime = _document != null ? _document.lastModified() : Long.MIN_VALUE;
        if (modTime > _loadModTime) {

            _loadModTime = modTime;

            final String content = _document.loadContent(getContext());
            if (!_document.isContentSame(_hlEditor.getText())) {

                final int[] sel = StringUtils.getSelection(_hlEditor);
                sel[0] = Math.min(sel[0], content.length());
                sel[1] = Math.min(sel[1], content.length());

                _hlEditor.withAutoFormatDisabled(() -> _hlEditor.setText(content));

                _hlEditor.setSelection(sel[0], sel[1]);
                StringUtils.showSelection(_hlEditor);
            }

            checkTextChangeState();

            if (_isPreviewVisible) {
                updateViewModeText();
            }
        }
    }

    @Override
    public boolean onOptionsItemSelected(@NonNull final MenuItem item) {
        final Activity activity = getActivity();
        if (activity == null) {
            return true;
        }
        _shareUtil.setContext(activity);
        final int itemId = item.getItemId();
        switch (itemId) {
            case R.id.action_undo: {
                if (_editTextUndoRedoHelper.getCanUndo()) {
                    _hlEditor.withAutoFormatDisabled(_editTextUndoRedoHelper::undo);
                    updateUndoRedoIconStates();
                }
                return true;
            }
            case R.id.action_redo: {
                if (_editTextUndoRedoHelper.getCanRedo()) {
                    _hlEditor.withAutoFormatDisabled(_editTextUndoRedoHelper::redo);
                    updateUndoRedoIconStates();
                }
                return true;
            }
            case R.id.action_save: {
                saveDocument(true);
                return true;
            }
            case R.id.action_reload: {
                final long oldModTime = _loadModTime;
                loadDocument();
                // Use modtime to show toast if document updated
                if (_loadModTime != oldModTime) {
                    Toast.makeText(activity, "✔", Toast.LENGTH_SHORT).show();
                }
                return true;
            }
            case R.id.action_preview: {
                setViewModeVisibility(true);
                return true;
            }
            case R.id.action_edit: {
                setViewModeVisibility(false);
                return true;
            }
            case R.id.action_preview_edit_toggle: {
                setViewModeVisibility(!_isPreviewVisible);
                return true;
            }
            case R.id.action_share_text: {
                if (saveDocument(false)) {
                    _shareUtil.shareText(_hlEditor.getText().toString(), "text/plain");
                }
                return true;
            }
            case R.id.action_share_file: {
                if (saveDocument(false)) {
                    _shareUtil.shareStream(_document.getFile(), "text/plain");
                }
                return true;
            }
            case R.id.action_share_html:
            case R.id.action_share_html_source: {
                if (saveDocument(false)) {
                    TextConverter converter = TextFormat.getFormat(_document.getFormat(), activity, _document).getConverter();
                    _shareUtil.shareText(converter.convertMarkup(_hlEditor.getText().toString(), _hlEditor.getContext(), false, _document.getFile()),
                            "text/" + (item.getItemId() == R.id.action_share_html ? "html" : "plain"));
                }
                return true;
            }
            case R.id.action_share_calendar_event: {
                if (saveDocument(false)) {
                    if (!_shareUtil.createCalendarAppointment(_document.getTitle(), _hlEditor.getText().toString(), null)) {
                        Toast.makeText(activity, R.string.no_calendar_app_is_installed, Toast.LENGTH_SHORT).show();
                    }
                }
                return true;
            }
            case R.id.action_share_screenshot:
            case R.id.action_share_image:
            case R.id.action_share_pdf: {
                _appSettings.getSetWebViewFulldrawing(true);
                if (saveDocument(false)) {
                    _nextConvertToPrintMode = true;
                    setViewModeVisibility(true);
                    Toast.makeText(activity, R.string.please_wait, Toast.LENGTH_LONG).show();
                    _webView.postDelayed(() -> {
                        if (item.getItemId() == R.id.action_share_pdf && Build.VERSION.SDK_INT >= Build.VERSION_CODES.KITKAT) {
                            _shareUtil.printOrCreatePdfFromWebview(_webView, _document, _hlEditor.getText().toString().contains("beamer\n"));
                        } else if (item.getItemId() != R.id.action_share_pdf) {
                            _shareUtil.shareImage(net.gsantner.opoc.util.ShareUtil.getBitmapFromWebView(_webView, item.getItemId() == R.id.action_share_image));
                        }
                    }, 7000);
                }

                return true;
            }
            case R.string.action_format_zimwiki:
            case R.string.action_format_keyvalue:
            case R.string.action_format_todotxt:
            case R.string.action_format_plaintext:
            case R.string.action_format_markdown: {
                if (_document != null) {
                    _document.setFormat(itemId);
                    applyTextFormat(itemId);
                    _appSettings.setDocumentFormat(_document.getPath(), _document.getFormat());
                }
                return true;
            }
            case R.id.action_search: {
                setViewModeVisibility(false);
                _textFormat.getTextActions().onSearch();
                return true;
            }
            case R.id.action_send_debug_log: {
                final String text = AppSettings.getDebugLog() + "\n\n------------------------\n\n\n\n" + Document.getMaskedContent(_hlEditor.getText().toString());
                _shareUtil.draftEmail("Debug Log " + getString(R.string.app_name_real), text, "debug@localhost.lan");
                return true;
            }

            case R.id.action_attach_color: {
                _textFormat.getTextActions().showColorPickerDialog();
                return true;
            }
            case R.id.action_attach_date: {
                DatetimeFormatDialog.showDatetimeFormatDialog(activity, _hlEditor);
                return true;
            }
            case R.id.action_attach_audio:
            case R.id.action_attach_file:
            case R.id.action_attach_image:
            case R.id.action_attach_link: {
                int actionId = (itemId == R.id.action_attach_audio ? 4 : (itemId == R.id.action_attach_image ? 2 : 3));
                AttachImageOrLinkDialog.showInsertImageOrLinkDialog(actionId, _document.getFormat(), activity, _hlEditor, _document.getFile());
                return true;
            }

            case R.id.action_load_epub: {
                FilesystemViewerCreator.showFileDialog(new FilesystemViewerData.SelectionListenerAdapter() {
                                                           @Override
                                                           public void onFsViewerSelected(String request, File file, final Integer lineNumber) {
                                                               _hlEditor.setText(CoolExperimentalStuff.convertEpubToText(file, getString(R.string.page)));
                                                           }

                                                           @Override
                                                           public void onFsViewerConfig(FilesystemViewerData.Options dopt) {
                                                               dopt.titleText = R.string.select;
                                                           }
                                                       }, getFragmentManager(), activity,
                        input -> input != null && input.getAbsolutePath().toLowerCase().endsWith(".epub")
                );
                return true;
            }
            case R.id.action_speed_read: {
                CoolExperimentalStuff.showSpeedReadDialog(activity, _hlEditor.getText().toString());
                return true;
            }
            case R.id.action_wrap_words: {
                wrapText = !wrapText;
                wrapTextSetting = wrapText;
                _appSettings.setDocumentWrapState(_document.getPath(), wrapTextSetting);
                setHorizontalScrollMode(wrapText);
                updateMenuToggleStates(0);
                return true;
            }
            case R.id.action_enable_highlighting: {
                highlightText = !highlightText;
                _hlEditor.setHighlightingEnabled(highlightText);
                _appSettings.setDocumentHighlightState(_document.getPath(), highlightText);
                updateMenuToggleStates(0);
                return true;
            }
            case R.id.action_enable_auto_format: {
                autoFormat = !autoFormat;
                _hlEditor.setAutoFormatEnabled(autoFormat);
                _appSettings.setDocumentAutoFormatEnabled(_document.getPath(), autoFormat);
                updateMenuToggleStates(0);
                return true;
            }
            case R.id.action_info: {
                if (_document != null) {
                    saveDocument(false); // In order to have the correct info displayed
                    FileInfoDialog.show(_document.getFile(), getFragmentManager());
                }
                return true;
            }
            case R.id.action_set_font_size: {
                SearchOrCustomTextDialogCreator.showFontSizeDialog(activity, _appSettings.getDocumentFontSize(_document.getPath()), (newSize) -> {
                    _hlEditor.setTextSize(TypedValue.COMPLEX_UNIT_SP, (float) newSize);
                    _appSettings.setDocumentFontSize(_document.getPath(), newSize);
                });
            }
            default: {
                return super.onOptionsItemSelected(item);
            }
        }
    }

    public void checkTextChangeState() {
        final boolean isTextChanged = !_document.isContentSame(_hlEditor.getText());

        if (_saveMenuItem != null && _saveMenuItem.isEnabled() != isTextChanged) {
            _saveMenuItem.setEnabled(isTextChanged).getIcon().mutate().setAlpha(isTextChanged ? 255 : 40);
        }
    }

    public void applyTextFormat(final int textFormatId) {
        final Activity activity = getActivity();
        if (activity == null) {
            return;
        }
        _textActionsBar.removeAllViews();
        _textFormat = TextFormat.getFormat(textFormatId, activity, _document);
        _hlEditor.setHighlighter(_textFormat.getHighlighter());
        _hlEditor.setDynamicHighlightingEnabled(_appSettings.isDynamicHighlightingEnabled());
        _hlEditor.setAutoFormatters(_textFormat.getAutoFormatInputFilter(), _textFormat.getAutoFormatTextWatcher());
        _hlEditor.setAutoFormatEnabled(_appSettings.getDocumentAutoFormatEnabled(_document.getPath()));
        _textFormat.getTextActions().setUiReferences(activity, _hlEditor, _webView);
        updateMenuToggleStates(textFormatId);
    }

    private void initDocState() {
        wrapTextSetting = _appSettings.getDocumentWrapState(_document.getPath());
        wrapText = isDisplayedAtMainActivity() || wrapTextSetting;

        highlightText = _appSettings.getDocumentHighlightState(_document.getPath(), _hlEditor.getText());
        autoFormat = _appSettings.getDocumentAutoFormatEnabled(_document.getPath());
        updateMenuToggleStates(0);

        setHorizontalScrollMode(wrapText);
        _hlEditor.setHighlightingEnabled(highlightText);
    }

    private void updateMenuToggleStates(final int selectedFormatActionId) {
        MenuItem mi;
        SubMenu su;
        if ((mi = _fragmentMenu.findItem(R.id.action_wrap_words)) != null) {
            mi.setChecked(wrapText);
        }
        if ((mi = _fragmentMenu.findItem(R.id.action_enable_highlighting)) != null) {
            mi.setChecked(highlightText);
        }
        if ((mi = _fragmentMenu.findItem(R.id.action_enable_auto_format)) != null) {
            mi.setChecked(autoFormat);
        }

        if (selectedFormatActionId != 0 && (mi = _fragmentMenu.findItem(R.id.submenu_format_selection)) != null && (su = mi.getSubMenu()) != null) {
            for (int i = 0; i < su.size(); i++) {
                if ((mi = su.getItem(i)).getItemId() == selectedFormatActionId) {
                    mi.setChecked(true);
                }
            }
        }
    }

    private void setHorizontalScrollMode(final boolean wrap) {

        final Context context = getContext();
        final boolean isCurrentlyWrap = _hsView == null || (_hlEditor.getParent() == _primaryScrollView);
        if (context != null && _hlEditor != null && isCurrentlyWrap != wrap) {

            final int[] sel = StringUtils.getSelection(_hlEditor);

            _primaryScrollView.removeAllViews();
            if (_hsView != null) {
                _hsView.removeAllViews();
            }
            if (!wrap) {
                if (_hsView == null) {
                    _hsView = new HorizontalScrollView(context);
                    _hsView.setFillViewport(true);
                }
                _hsView.addView(_hlEditor);
                _primaryScrollView.addView(_hsView);
            } else {
                _primaryScrollView.addView(_hlEditor);
            }

            // Run after layout() of immediate parent completes
            (wrap ? _primaryScrollView : _hsView).post(() -> StringUtils.setSelectionAndShow(_hlEditor, sel[0], sel[1]));
        }
    }

    @Override
    public String getFragmentTag() {
        return FRAGMENT_TAG;
    }

    public boolean checkPermissions() {

        if (_document != null) {
            final File file = _document.getFile();

            if (_shareUtil.isUnderStorageAccessFolder(file, false) && _shareUtil.getStorageAccessFrameworkTreeUri() == null) {
                _shareUtil.showMountSdDialog(getActivity());
            }

            final boolean permok = _document.testCreateParent() && _shareUtil.canWriteFile(file, false, true);
            _textSdWarning.setVisibility(permok ? View.GONE : View.VISIBLE);
            return permok;
        }
        return false;
    }

    // Save the file
    // Only supports java.io.File. TODO: Android Content
    public boolean saveDocument(final boolean forceSaveEmpty) {
        // Document is written iff writeable && content has changed
        final CharSequence text = _hlEditor.getText();
        if (_document != null && !_document.isContentSame(text) && checkPermissions() && isAdded()) {
            if (_document.saveContent(getActivity(), text, _shareUtil, forceSaveEmpty)) {
                checkTextChangeState();
                return true;
            } else {
                return false; // Failure only if saveContent somehow fails
            }
        } else {
            return true; // Report success if text not changed
        }
    }

    private boolean isDisplayedAtMainActivity() {
        return getActivity() instanceof MainActivity;
    }

    public void updateViewModeText() {
        _textFormat.getConverter().convertMarkupShowInWebView(_document, _hlEditor.getText().toString(), _webView, _nextConvertToPrintMode);
    }

    public void setViewModeVisibility(final boolean show) {
        final Activity activity = getActivity();
        _textFormat.getTextActions().recreateTextActionBarButtons(_textActionsBar, show ? TextActions.ActionItem.DisplayMode.VIEW : TextActions.ActionItem.DisplayMode.EDIT);
        if (show) {
            updateViewModeText();
            new ActivityUtils(activity).hideSoftKeyboard().freeContextRef();
            _hlEditor.clearFocus();
            _hlEditor.postDelayed(() -> new ActivityUtils(activity).hideSoftKeyboard().freeContextRef(), 300);
            fadeInOut(_webView, _primaryScrollView);
        } else {
            _webViewClient.setRestoreScrollY(_webView.getScrollY());
            fadeInOut(_primaryScrollView, _webView);
        }

        _nextConvertToPrintMode = false;
        _isPreviewVisible = show;

        ((AppCompatActivity) activity).supportInvalidateOptionsMenu();
    }

    // Callback from view-mode/javascript
    @SuppressWarnings("unused")
    @JavascriptInterface
    public void webViewJavascriptCallback(final String[] jsArgs) {
        final String[] args = (jsArgs == null || jsArgs.length == 0 || jsArgs[0] == null) ? new String[0] : jsArgs;
        final String type = args.length == 0 || TextUtils.isEmpty(args[0]) ? "" : args[0];
    }

    private static boolean fadeInOut(final View in, final View out) {
        // Do nothing if we are already in the correct state
        if (in.getVisibility() == View.VISIBLE && out.getVisibility() == View.GONE) {
            return false;
        }

        in.setAlpha(0);
        in.setVisibility(View.VISIBLE);
        in.animate().alpha(1).setDuration(200).setListener(null);
        out.animate()
                .alpha(0)
                .setDuration(200)
                .setListener(new AnimatorListenerAdapter() {
                    @Override
                    public void onAnimationEnd(Animator animation) {
                        out.setVisibility(View.GONE);
                    }
                });

        return true;
    }

    @Override
    protected boolean onToolbarLongClicked(View v) {
        if (getUserVisibleHint()) {
            _textFormat.getTextActions().runJumpBottomTopAction(_isPreviewVisible ? TextActions.ActionItem.DisplayMode.VIEW : TextActions.ActionItem.DisplayMode.EDIT);
            return true;
        }
        return false;
    }

    //
    //
    //
    //

    public Document getDocument() {
        return _document;
    }

    public WebView getWebview() {
        return _webView;
    }

    @Override
    protected void onToolbarClicked(View v) {
        if (!_isPreviewVisible && _textFormat != null) {
            _textFormat.getTextActions().runTitleClick();
        }
    }

    @Override
    public void onDestroy() {
        try {
            _webView.loadUrl("about:blank");
            _webView.destroy();
        } catch (Exception ignored) {
        }
        super.onDestroy();
    }
}
